use std::net::{IpAddr, Ipv4Addr};

#[derive(serde::Serialize, serde::Deserialize, Clone, Copy, Debug)]
pub struct XMLUuid(pub uuid::Uuid);

impl XMLUuid {
    pub fn parse_from_str(s: &str) -> anyhow::Result<Self> {
        Ok(Self(uuid::Uuid::parse_str(s)?))
    }

    pub fn new_random() -> Self {
        Self(uuid::Uuid::new_v4())
    }
    pub fn as_string(&self) -> String {
        self.0.to_string()
    }

    pub fn is_valid(&self) -> bool {
        log::debug!("UUID version ({}): {}", self.0, self.0.get_version_num());
        self.0.get_version_num() == 4
    }
}

/// OS information
#[derive(serde::Serialize, serde::Deserialize)]
#[serde(rename = "os")]
pub struct OSXML {
    #[serde(rename(serialize = "@firmware"), default)]
    pub firmware: String,
    pub r#type: OSTypeXML,
    pub loader: Option<OSLoaderXML>,
}

/// OS Type information
#[derive(serde::Serialize, serde::Deserialize)]
#[serde(rename = "os")]
pub struct OSTypeXML {
    #[serde(rename(serialize = "@arch"))]
    pub arch: String,
    #[serde(rename(serialize = "@machine"))]
    pub machine: String,
    #[serde(rename = "$value")]
    pub body: String,
}

/// OS Loader information
#[derive(serde::Serialize, serde::Deserialize)]
#[serde(rename = "loader")]
pub struct OSLoaderXML {
    #[serde(rename(serialize = "@secure"))]
    pub secure: String,
}

/// Hypervisor features
#[derive(serde::Serialize, serde::Deserialize, Default)]
#[serde(rename = "features")]
pub struct FeaturesXML {
    pub acpi: ACPIXML,
}

/// ACPI feature
#[derive(serde::Serialize, serde::Deserialize, Default)]
#[serde(rename = "acpi")]
pub struct ACPIXML {}

/// Devices information
#[derive(serde::Serialize, serde::Deserialize)]
#[serde(rename = "devices")]
pub struct DevicesXML {
    /// Graphics (used for VNC
    #[serde(skip_serializing_if = "Option::is_none")]
    pub graphics: Option<GraphicsXML>,

    /// Disks (used for storage)
    #[serde(default, rename = "disk", skip_serializing_if = "Vec::is_empty")]
    pub disks: Vec<DiskXML>,
}

/// Screen information
#[derive(serde::Serialize, serde::Deserialize)]
#[serde(rename = "graphics")]
pub struct GraphicsXML {
    #[serde(rename(serialize = "@type"))]
    pub r#type: String,
    #[serde(rename(serialize = "@socket"))]
    pub socket: String,
}

/// Disk information
#[derive(serde::Serialize, serde::Deserialize)]
#[serde(rename = "disk")]
pub struct DiskXML {
    #[serde(rename(serialize = "@type"))]
    pub r#type: String,
    #[serde(rename(serialize = "@device"))]
    pub r#device: String,

    pub driver: DiskDriverXML,
    pub source: DiskSourceXML,
    pub target: DiskTargetXML,
    #[serde(skip_serializing_if = "Option::is_none")]
    pub readonly: Option<DiskReadOnlyXML>,
    pub boot: DiskBootXML,
    #[serde(skip_serializing_if = "Option::is_none")]
    pub address: Option<DiskAddressXML>,
}

#[derive(serde::Serialize, serde::Deserialize)]
#[serde(rename = "driver")]
pub struct DiskDriverXML {
    #[serde(rename(serialize = "@name"))]
    pub name: String,
    #[serde(rename(serialize = "@type"))]
    pub r#type: String,
    #[serde(default, rename(serialize = "@cache"))]
    pub r#cache: String,
}

#[derive(serde::Serialize, serde::Deserialize)]
#[serde(rename = "source")]
pub struct DiskSourceXML {
    #[serde(rename(serialize = "@file"))]
    pub file: String,
}

#[derive(serde::Serialize, serde::Deserialize)]
#[serde(rename = "target")]
pub struct DiskTargetXML {
    #[serde(rename(serialize = "@dev"))]
    pub dev: String,
    #[serde(rename(serialize = "@bus"))]
    pub bus: String,
}

#[derive(serde::Serialize, serde::Deserialize)]
#[serde(rename = "readonly")]
pub struct DiskReadOnlyXML {}

#[derive(serde::Serialize, serde::Deserialize)]
#[serde(rename = "boot")]
pub struct DiskBootXML {
    #[serde(rename(serialize = "@order"))]
    pub order: String,
}

#[derive(serde::Serialize, serde::Deserialize)]
#[serde(rename = "address")]
pub struct DiskAddressXML {
    #[serde(rename(serialize = "@type"))]
    pub r#type: String,
    #[serde(
        default,
        skip_serializing_if = "Option::is_none",
        rename(serialize = "@controller")
    )]
    pub r#controller: Option<String>,
    #[serde(rename(serialize = "@bus"))]
    pub r#bus: String,
    #[serde(
        default,
        skip_serializing_if = "Option::is_none",
        rename(serialize = "@target")
    )]
    pub r#target: Option<String>,
    #[serde(
        default,
        skip_serializing_if = "Option::is_none",
        rename(serialize = "@unit")
    )]
    pub r#unit: Option<String>,
}

/// Domain RAM information
#[derive(serde::Serialize, serde::Deserialize)]
#[serde(rename = "memory")]
pub struct DomainMemoryXML {
    #[serde(rename(serialize = "@unit"))]
    pub unit: String,

    #[serde(rename = "$value")]
    pub memory: usize,
}

#[derive(serde::Serialize, serde::Deserialize)]
#[serde(rename = "topology")]
pub struct DomainCPUTopology {
    #[serde(rename(serialize = "@sockets"))]
    pub sockets: usize,
    #[serde(rename(serialize = "@cores"))]
    pub cores: usize,
    #[serde(rename(serialize = "@threads"))]
    pub threads: usize,
}

#[derive(serde::Serialize, serde::Deserialize)]
#[serde(rename = "cpu")]
pub struct DomainCPUXML {
    #[serde(rename(serialize = "@mode"))]
    pub mode: String,
    pub topology: Option<DomainCPUTopology>,
}

/// Domain information, see https://libvirt.org/formatdomain.html
#[derive(serde::Serialize, serde::Deserialize)]
#[serde(rename = "domain")]
pub struct DomainXML {
    /// Domain type (kvm)
    #[serde(rename(serialize = "@type"))]
    pub r#type: String,

    pub name: String,
    pub uuid: Option<XMLUuid>,
    pub genid: Option<uuid::Uuid>,
    pub title: Option<String>,
    pub description: Option<String>,
    pub os: OSXML,
    #[serde(default)]
    pub features: FeaturesXML,
    pub devices: DevicesXML,

    /// The maximum allocation of memory for the guest at boot time
    pub memory: DomainMemoryXML,

    /// CPU information
    pub cpu: DomainCPUXML,

    pub on_poweroff: String,
    pub on_reboot: String,
    pub on_crash: String,
}

/// Domain state
#[derive(serde::Serialize, Debug, Copy, Clone)]
pub enum DomainState {
    NoState,
    Running,
    Blocked,
    Paused,
    Shutdown,
    Shutoff,
    Crashed,
    PowerManagementSuspended,
    Other,
}

/// Network forward information
#[derive(serde::Serialize, serde::Deserialize, Debug)]
#[serde(rename = "forward")]
pub struct NetworkForwardXML {
    #[serde(rename(serialize = "@mode"))]
    pub mode: String,
    #[serde(
        default,
        rename(serialize = "@dev"),
        skip_serializing_if = "String::is_empty"
    )]
    pub dev: String,
}

/// Network DNS information
#[derive(serde::Serialize, serde::Deserialize, Debug)]
#[serde(rename = "dns")]
pub struct NetworkDNSXML {
    pub forwarder: NetworkDNSForwarderXML,
}

/// Network DNS information
#[derive(serde::Serialize, serde::Deserialize, Debug)]
#[serde(rename = "fowarder")]
pub struct NetworkDNSForwarderXML {
    /// Address of the DNS server
    #[serde(rename(serialize = "@addr"))]
    pub addr: Ipv4Addr,
}

/// Network DNS information
#[derive(serde::Serialize, serde::Deserialize, Debug)]
#[serde(rename = "domain")]
pub struct NetworkDomainXML {
    #[serde(rename(serialize = "@name"))]
    pub name: String,
}

fn invalid_prefix() -> u32 {
    u32::MAX
}

fn invalid_ip() -> IpAddr {
    IpAddr::V4(Ipv4Addr::BROADCAST)
}

/// Network ip information
#[derive(serde::Serialize, serde::Deserialize, Debug)]
#[serde(rename = "ip")]
pub struct NetworkIPXML {
    #[serde(default, rename(serialize = "@family"))]
    pub family: String,
    #[serde(rename(serialize = "@address"))]
    pub address: IpAddr,
    /// Network Prefix
    #[serde(rename(serialize = "@prefix"), default = "invalid_prefix")]
    pub prefix: u32,
    /// Network Netmask. This field is never serialized, but because we can't know if LibVirt will
    /// provide us netmask or prefix, we need to handle both of these fields
    #[serde(
        rename(serialize = "@netmask"),
        default = "invalid_ip",
        skip_serializing
    )]
    pub netmask: IpAddr,
    pub dhcp: Option<NetworkDHCPXML>,
}

#[derive(serde::Serialize, serde::Deserialize, Debug)]
#[serde(rename = "dhcp")]
pub struct NetworkDHCPXML {
    pub range: NetworkDHCPRangeXML,
}

#[derive(serde::Serialize, serde::Deserialize, Debug)]
#[serde(rename = "dhcp")]
pub struct NetworkDHCPRangeXML {
    #[serde(rename(serialize = "@start"))]
    pub start: IpAddr,
    #[serde(rename(serialize = "@end"))]
    pub end: IpAddr,
}

/// Network information, see https://libvirt.org/formatnetwork.html
#[derive(serde::Serialize, serde::Deserialize, Debug)]
#[serde(rename = "network")]
pub struct NetworkXML {
    pub name: String,
    #[serde(skip_serializing_if = "Option::is_none")]
    pub uuid: Option<XMLUuid>,
    #[serde(skip_serializing_if = "Option::is_none")]
    pub title: Option<String>,
    #[serde(skip_serializing_if = "Option::is_none")]
    pub description: Option<String>,
    #[serde(skip_serializing_if = "Option::is_none")]
    pub forward: Option<NetworkForwardXML>,
    #[serde(skip_serializing_if = "Option::is_none")]
    pub dns: Option<NetworkDNSXML>,
    #[serde(skip_serializing_if = "Option::is_none")]
    pub domain: Option<NetworkDomainXML>,
    #[serde(default, rename = "ip")]
    pub ips: Vec<NetworkIPXML>,
}
